/**
* Knoten in DOM-Baum einfügen
* @return Element   Erzeugter Knoten
*/
function createElementNS(
    strNamespace,   // Namespace
    strTagname,     // Optional: Tagname des zu erzeugenden Elementes
    nodeParent,     // Optional: Künftiger Elternknoten
    objAtts,        // Optional: Attribute des zu erzeugenden Elementes
    strTextContent, // Optional: Textinhalt des Elementes
    nodeBefore      // Optional: Knoten, als deren previousSibling der neue Knoten eingefügt werden soll
) {
    var node = nodeBefore ? nodeBefore.parentNode : nodeParent;
    if (strTagname)
        node = (nodeParent ? nodeParent.ownerDocument : document).createElementNS(strNamespace, strTagname);
    if (objAtts)
        for (var key in objAtts)
            node.setAttribute(key, objAtts[key]);
    if ((strTextContent !== undefined) && ('' + strTextContent).length)
        node.appendChild((nodeParent ? nodeParent.ownerDocument : document).createTextNode(strTextContent));
    if (nodeParent && (node != nodeParent))
        nodeParent.insertBefore(node, nodeBefore);
    return node;
}

/**
* HTML-Element in DOM-Baum einfügen
* @return Element   Erzeugter Knoten
*/
function createHtml(
    strTagname,     // Optional: Tagname des zu erzeugenden Elementes
    nodeParent,     // Optional: Künftiger Elternknoten
    objAtts,        // Optional: Attribute des zu erzeugenden Elementes
    strTextContent, // Optional: Textinhalt des Elementes
    nodeBefore      // Optional: Knoten, als deren previousSibling der neue Knoten eingefügt werden soll
) {
    return createElementNS(
        'htt' + 'p://w' + 'ww.w3.org/1999/xhtml',
        strTagname,
        nodeParent,
        objAtts,
        strTextContent,
        nodeBefore
    );
}

/**
* SVG-Element in DOM-Baum einfügen
* @return Element   Erzeugter Knoten
*/
function createSvg(
    strTagname,     // Optional: Tagname des zu erzeugenden Elementes
    nodeParent,     // Optional: Künftiger Elternknoten
    objAtts,        // Optional: Attribute des zu erzeugenden Elementes
    strTextContent, // Optional: Textinhalt des Elementes
    nodeBefore      // Optional: Knoten, als deren previousSibling der neue Knoten eingefügt werden soll
) {
    return createElementNS(
        'htt' + 'p://w' + 'ww.w3.org/2000/svg',
        strTagname,
        nodeParent,
        objAtts,
        strTextContent,
        nodeBefore
    );
}

/**
* Klasse, die Schritt-für-Schritt-Anleitung steuert
*/
function StepByStep(
    body,           // DOM-Node des Container-Elements
    lessonData,     // Objekt mit Informationen zur Unterrichtsstunde
    desc,           // Menschenlesbare Beschreibung der Inhaltsart
    contentClass    // Klasse, welche die Vorführung des Inhalts verwalten soll
) {
    this.parentNode = body;
    body.innerHTML = '';
    createHtml('h1', body, null, desc + ' wird geladen…');
    createHtml('p', body, null, 'Dies kann je nach Internetverbindung und Größe der nötigen Dateien einige Zeit in Anspruch nehmen…');
    this.content = new contentClass(this, lessonData);
    this._loadAudio(lessonData);
}

// Ist diese Instanz einsatzbereit?
StepByStep.prototype.ready = false;

// Container, in dem sich Steuerungselemente und Bildschirmpräsentation befinden
StepByStep.prototype.parentNode = null;

// Zuletzt erstellte Meldung
StepByStep.prototype._lastMsgNode = null;

// Inhalt, welche die Vorführung des Inhalts verwaltet
StepByStep.prototype.content = null;

/**
* Eine Erfolgs- oder Fehlermeldung anzeigen
* @return undefined
*/
StepByStep.prototype.showMsg = function(
    strMsg,         // Erfolgs- oder Fehlermeldung
    boolSuccess     // Optional: Handelt es sich um eine Erfolgsmeldung? (Standardwert: false)
) {
    if (this._lastMsgNode && this._lastMsgNode.parentNode)
        this._lastMsgNode.parentNode.removeChild(this._lastMsgNode);
    var nodeBefore = this.parentNode.firstChild;
    if (nodeBefore && (nodeBefore.nodeType == Node.ELEMENT_NODE) && (nodeBefore.nodeName.toUpperCase() == 'H1'))
        nodeBefore = nodeBefore.nextSibling;
    this._lastMsgNode = createHtml('p', this.parentNode, { 'class': boolSuccess ? 'msg_success' : 'msg_failed' }, undefined, nodeBefore);
    createHtml('span', createHtml('span', this._lastMsgNode), null, strMsg);
    if (typeof this._lastMsgNode.scrollIntoView == 'function')
        this._lastMsgNode.scrollIntoView({ behavior: 'smooth' });
}

/**
* Die benötigten Audiodateien laden und die interne Auflistung der Schritte erstellen
* @return undefined
*/
StepByStep.prototype._loadAudio = function(
    lessonData  // Objekt mit Informationen zur Unterrichtsstunde
) {
    this.steps = [];
    if (!(lessonData.steps instanceof Array))
        return;
    var step;
    const basedir = lessonData.basedir ? (lessonData.basedir + '/') : '';
    for (var z = 0; z < lessonData.steps.length; ++z) {
        step = { svg: null };
        if (lessonData.steps[z].audio) {
            step.audio = new Audio(basedir + lessonData.steps[z].audio);
            step.audio.controls = true;
        }
        this.steps.push(step);
    }
}

/**
* Dem LMS mitteilen, dass die Bearbeitung jetzt los geht und ggf. vorherige Werte laden
* @return undefined
*/
StepByStep.prototype._LMSconnect = function() {
    if (this.lms || (typeof LMSLink != 'function'))
        return;
    Object.defineProperty(this, 'lms', { writeable: false, value: new LMSLink() });
    var self = this;
    this.lms.onnew = function() {
        this.score = 0;
    }
    this.lms.onresume = function() {
        self.done = Math.min(Math.max(0, this.score), self.steps.length);
        if (!Number.isFinite(self.done))
            self.done = 0;
        if (self.done)
            self.setActive((self.done < self.steps.length) ? self.done : 0, true);
    }
    this.lms.start();
    if (this.lms.lessonStatus == 'not attempted')
        this.lms.lessonStatus = 'incomplete';
    this.lms.scoreMin = 0;
    this.lms.scoreMax = this.steps.length;
    if (this.lms.ready) {
        var mode = this.lms.lessonMode;
        if ((mode == 'browse') || (mode == 'review')) {
            this.done = self.steps.length;
            this.setActive(0);
        }
    } else
        this.showMsg('Achtung: Dein Fortschritt kann nicht gespeichert werden, weil keine Verbindung zur Lernplattform hergestellt werden konnte.');
}

/**
* Navigationsleite erstellen
* @return undefined
*/
StepByStep.prototype._createNav = function(
    body    // DOM-Node des Body-Elements
) {
    var self = this;
    var navUl = createHtml('ul', createHtml('nav', body), { id: 'controls' });
    this._nav = navUl;
    // zurück
    this.prevBtn = createHtml('a', createHtml('li', navUl), { href: '#' }, '← zurück');
    this.prevBtn.addEventListener('click', function (ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.setActive(self.active - 1);
    });

    // aktueller Schritt
    var nodeLi = createHtml('form', createHtml('li', navUl), { action: '#', method: 'post' }, 'Schritt ');
    this.inputStep = createHtml('input', nodeLi, { type: 'number', min: 1, max: this.steps.length, step: 1, required: 'required', size: 2 });
    createHtml(undefined, nodeLi, undefined, ' von ' + this.steps.length);
    this.inputStep.addEventListener('change', function() {
        self.setActive(self.inputStep.value);
    });
    nodeLi.addEventListener('submit', function (ev) {
        ev.preventDefault();
        ev.stopPropagation();
    });

    // Ort für Audio-Element
    this.audioPlace = createHtml('li', navUl);

    // weiter
    this.nextBtn = createHtml('a', createHtml('li', navUl), { href: '#' }, 'weiter →');
    this.nextBtn.addEventListener('click', function (ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.setActive(self.active + 1);
    });
    body.insertBefore(navUl.parentNode, body.firstChild);
}

/**
* Lautstärke bei allen Audio-Elementen einstellen
* @return undefined
*/
StepByStep.prototype.setVolume = function(
    volume  // Neue Lautstärke
) {
    this._volume = volume;
    for (var z = 0; z < this.steps.length; ++z)
        if (this.steps[z].audio)
            this.steps[z].audio.volume = volume
}

/**
* Wiedergabegeschwindigkeit bei allen Audio-Elementen einstellen
* @return undefined
*/
StepByStep.prototype.setPlaybackRate = function(
    playbackRate    // Neue Wiedergabegeschwindigkeit
) {
    this._playbackRate = playbackRate;
    for (var z = 0; z < this.steps.length; ++z)
        if (this.steps[z].audio)
            this.steps[z].audio.playbackRate = playbackRate
}

/**
* Navigationsleiste ggf. an zu geringe Breite anpassen
* @return undefined
*/
StepByStep.prototype.rearrangeNav = function() {
    if (!this.audio) return;

    // Herausfinden, wie breit Audioelemente sind
    var audioWidth = this.audio.clientWidth;

    // Bestimmen, ob Navigationsleiste mehrzeilig sein soll
    var boolMehrzeilig = (1.8 * audioWidth >= window.innerWidth);

    // Entsprechende Änderungen vornehmen, falls nötig
    if (boolMehrzeilig) {
        if (this._secondNav) return;
        var audioHeight = this.audio.clientHeight;
        this._nav.parentNode.setAttribute('style', 'margin-bottom:' + audioHeight + 'px');
        this._secondNav = createHtml('ul', this._nav.parentNode);
        this._secondNav.appendChild(this.audioPlace);
    } else {
        if (!this._secondNav) return;
        this._nav.parentNode.removeAttribute('style');
        this._nav.insertBefore(this.audioPlace, this._nav.lastChild);
        this._secondNav.parentNode.removeChild(this._secondNav);
        this._secondNav = null;
    }
}

/**
* Aktualisieren, welcher der Weiter-/Zurückschaltflächen anklickbar ist
* @return undefined
*/
StepByStep.prototype.updateBtnState = function() {
    var boolAllowNext = (this.active <= this.done);
    var step = Math.min(Math.max(this.active, 0), this.steps.length);
    var index = Math.max(0, step - 1);

    if (index)
        this.prevBtn.classList.remove('disabled');
    else
        this.prevBtn.classList.add('disabled');
    if (boolAllowNext)
        this.nextBtn.classList.remove('disabled');
    else
        this.nextBtn.classList.add('disabled');
}

/**
* Beschriftung der Weiterschaltfläche aktualisieren
* @return undefined
*/
StepByStep.prototype.updateBtnLabel = function() {
    var strNext = "weiter →";
    if (!this.active)
        strNext = "Starten →";
    else if (this.active == this.steps.length)
        strNext = "Fertig →";
    if (this.nextBtn.textContent != strNext)
        this.nextBtn.textContent = strNext;
}

/**
* Der Lernplattform mitteilen, wie lange schon am Lernpaket gearbeitet wurde, was der aktuelle Status ist usw.
* @return undefined
*/
StepByStep.prototype.updateLmsStatus = function() {
    if (!this.lms || !this.lms.ready)
        return;
    var mode = this.lms.lessonMode;
    if (mode == 'browse')
        return;
    this.lms.sessionTime = this.lms.sessionTime;
    if (mode == 'review')
        return;
    this.lms.score = this.done;
    var mastery = this.lms.scoreMastery;
    if ((this.done >= this.steps.length) || (mastery && (this.done > mastery)))
        this.lms.lessonStatus = 'completed';
}

/**
* Speichert einen neuen Wert für die Anzahl der abgeschlossenen Schritte, sofern der neue Wert größer als der alte ist
* @return undefined
*/
StepByStep.prototype.saveDone = function(
    value       // Neuer Wert für die Anzahl der abgeschlossenen Schritte
) {
    this.done = Math.min(Math.max(this.done, value), this.steps.length);
    this.updateLmsStatus();
}

/**
* Sicherstellen, dass aktueller Schritt aktiv ist
* @return undefined
*/
StepByStep.prototype.changeActiveTo = function(
    activeNew       // Nummer des Inhaltes, der von nun an aktiv sein soll
) {
    var indexOld = Number.isFinite(this.active) ? Math.max(0, this.active - 1) : -1;
    var step = Math.min(Math.max(activeNew, 0), this.steps.length);
    this.active = step;
    var index = Math.max(0, step - 1);
    var self = this;

    // Vor-/Zurück-Link
    this.updateBtnState();
    this.updateBtnLabel();

    // Schrittanzeige
    this.inputStep.value = index + 1;

    // Richtiges Audio-Element anzeigen
    if (this.audio)
        this.audio.pause();
    this.audio = null;
    while (this.audioPlace.firstChild)
        this.audioPlace.removeChild(this.audioPlace.firstChild);
    if (this.steps[index].audio) {
        this.audio = this.steps[index].audio.cloneNode(true);
        this.audio.addEventListener('volumechange', function() {
            self.setVolume(self.audio.volume);
        });
        this.audio.addEventListener('ratechange', function() {
            self.setPlaybackRate(self.audio.playbackRate);
        });
        this.audio.addEventListener('ended', function() {
            self.saveDone(index + 1);
            self.updateBtnState();
        });
        if (step == 0) {
            this.audio.addEventListener('play', function() {
                self.active = 1;
            });
        }
        if (this._volume !== undefined)
            this.audio.volume = this._volume;
        if (this._playbackRate !== undefined)
            this.audio.playbackRate = this._playbackRate;
        this.audioPlace.appendChild(this.audio);
    } else {
        self.saveDone(index + 1);
        self.updateBtnState();
    }

    // Richtigen Inhalt anzeigen
    this.content.changeActive(indexOld, index, this.steps.map(function(el) { return el.svg; }));

    // ggf. an zu kleinen Bildschirm anpassen
    this.rearrangeNav();

    // Aktive Ebene in den Sichtbereich scrollen
    this.content.scrollToActive(this.steps[index].svg, this._nav.parentNode.clientHeight);
}

/**
* Nummer des aktiven Schritts setzen und ggf. automatisch Audiowiedergabe starten
* @return undefined
*/
StepByStep.prototype.setActive = function(
    value,      // Nummer des neuen aktiven Schritts
    noAutoPlay  // Optional: Das automatische Abspielen der Audiodatei unterbinden?
) {
    var next = Math.min(Math.max(value, 0), this.steps.length + 1, this.done + 1);
    if (next > this.steps.length) {
        if (this.lms && this.lms.ready)  {
            this.lms.exit = '';
            var mode = this.lms.lessonMode;
            if ((mode != 'browse') && (mode != 'review'))
                this.lms.lessonStatus = 'completed';
            this.lms.stop();
        }
        this.showMsg('Du hast diese Online-Unterrichtsstunde erfolgreich absolviert!', true);
        this.active = this.steps.length;
        next = this.active;
    } else if(value > next) {
        this.showMsg('Höre dir erst Schritt ' + next + ' komplett an! Erst dann kannst du fortfahren.');
    }
    if (next != this.active) {
        this.changeActiveTo(next);
    } else if (this.audio && !this.audio.paused)
        return;
    if (this.active && this.audio && !noAutoPlay)
        this.audio.play();
}

// Informationen zu jedem Schritt der Form [ { audio: Node, svg: Node }, … ]
StepByStep.prototype.steps = [];

// Schaltfläche zum vorherigen Schritt
StepByStep.prototype.prevBtn = null;

// Schaltfläche zum nächsten Schritt
StepByStep.prototype.nextBtn = null;

// Eingabefeld für Nummer des aktiven Schrittes
StepByStep.prototype.inputStep = null;

// Ort, an dem das aktuelle Audio-Element sein soll
StepByStep.prototype.audioPlace = null;

// Navigationstabellenzeile
StepByStep.prototype._nav = null;

// Ggf. zweite Navigationstabellenzeile nur für das Audioelement
StepByStep.prototype._secondNav = null;

// Aktives Audio-Element
StepByStep.prototype.audio = null;

// Nummer des aktiven Schritts
StepByStep.prototype.active = -Infinity;

// Wie viele Schritte wurden bereits abgeschlossen?
StepByStep.prototype.done = 0;

/**
* Die Vorführung starten
* @return undefined
*/
StepByStep.prototype.establishContent = function(
    nodeContent,    // DOM-Element, welches als Inhalt angezeigt werden soll
    arrSteps        // Schritte aus Sicht des Inhalts
) {
    this.parentNode.innerHTML = '';
    this.parentNode.appendChild(nodeContent);
    this.ready = true;

    var z;
    for (z = 0; z < arrSteps.length; ++z) {
        if (z < this.steps.length)
            this.steps[z].content = arrSteps[z];
        else
            this.steps.push({ audio: null, content: arrSteps[z] });
    }
    this._createNav(this.parentNode);
    this.changeActiveTo(0);
    this._LMSconnect();
}
